{
                                                  Date: June 24, 1995
ͻ
 vga2pcx.zip package 
ͼ

                  *FAST* Turbo Pascal unit (part assembly language).


  DESCRIPTION:

  Saves a copy of the 640x480 VGA 16 color screen to a properly
  formatted PCX file.  This routine also takes care to save
  the current palette so your colors will be preserved correctly.


  REQUIRED:

  *  Turbo Pascal 6+ to run these routines, however, if you just want
     to learn the PCX format you can read the source code and comments.
     I think Pascal code is a relatively easy language to read and
     understand.


  COMMON USES:

  *  Programmer can give user ability to save a copy of his/her
     exact graphics screen for presentation or documentation purposes.
     The PCX format is common format that is easily imported into
     word processors, etc.

  *  Slide show programs often operate on PCX files.  You can save
     a sequence of pictures and then run a program to cycle through
     them like a demonstration.

  *  Frees the DOS programmer from having to support multiple
     printing devices.  If you can fit your graphics on the screen,
     the screen can be saved into a standard PCX file.  This opens
     the door to printing via other, more specialized, programs.

  *  Source code included, so you can learn the PCX format and
     customize for your own use.


  AUTHOR INFO:

  Bren Sessions
  1590 NW Maple
  Corvallis, OR  97330

  Tel: (503) 758-2256
  Email: sessioj@ece.orst.edu


  PARTICULARS:

  No copyright!!!   Feel free to incorporate this into your own code.

  If you find these functions useful, I'd appreciate to hear from you.
  If you feel like supporting my efforts, be great too.  Some people
  have sent me mail telling that they have been waiting for these
  routines for years.  I suggest $5, but you can send me whatever you
  think is reasonable.
}

unit vga2pcx;
{$R-,S-}

interface

type nt = string[12];

procedure write_pcx(fn : nt; var fl : boolean);

{
  Will save the current 640x480 VGA 16 color screen into file passed
  as parameter 'fn'.  If the save is successful (e.g. the filename
  was legal, 'fl' is given the value of true.
}

implementation

const BUFSIZE = 512;   { will write to disk at this byte interval }
  rm : array[0..15] of byte = (0,1,2,3,4,5,20,7,56,57,58,59,60,61,62,63);

type header_rec =
         record
           pcx_id  : byte;   { 0) 0x0a = ZSoft .PCX file          }
           pcx_ver : byte;   { 1) 0x05 = PC PaintBrush 3.0        }
           encode  : byte;   { 2) 0x01 = RLE                      }
           bpp     : byte;   { 3) 0x01 = bits/pixel on plane      }
           left    : word;   { 4-5) Window Left                   }
           top     : word;   { 6-7) Window Top                    }
           right   : word;   { 8-9) Window Right                  }
           bott    : word;   { 10-11) Window Bottom               }
           xres    : word;   { 12-13) Horizontal resolution       }
           yres    : word;   { 14-15) Vertical resolution         }
           rgb     : array[0..15,0..2] of byte;{ (R-G-B) values   }
           resv    : byte;   { 64) Reserved                       }
           bplanes : byte;   { 65) Number of bit planes, VGA16=4  }
           bpl     : word;   { 66-67) # of bytes/line, VGA16=80   }
           ptype   : word;   { 68-69) Palette type, color=1       }
           xsz     : word;
           ysz     : word;
           unused  : array[74..127] of byte;
         end;
     pal_rec = array[0..63,0..2] of byte;
     tbyte = ^byte;

var fz  : file;
    buf : array[1..BUFSIZE] of byte;
    bi  : integer;   { buffer index }

{
procedure get_rgb(color : integer; var r,g,b : integer);

begin

  r:=(((color and $20) shr 5) or ((color and $04) shr 1))*84;
  g:=(((color and $10) shr 4) or ((color and $02)      ))*84;
  b:=(((color and $08) shr 3) or ((color and $01) shl 1))*84;

end;
}

procedure get_pal(var pal : pal_rec); assembler;

asm

  mov ax,1201h   { disable refresh }
  mov bl,36h
  int 10h

  mov ax,1017h
  xor bx,bx
  mov cx,64

  les dx,pal
  int 10h

  mov ax,1200h   { enable refresh }
  mov bl,36h
  int 10h

end;

{-----------------------------------------------------------------------}

procedure get_bitplane_at_scan_line(plane,scanline : word; address : tbyte); assembler;

{
  Dumps the requested scanline (0-479) at a particular bitplane (0-3)
  into a memory address.  Space required = 80 bytes (640 pixel).
}

asm

  cld

  mov bx,ds

  mov ax,SegA000
  mov ds,ax

  mov ax,80
  mul scanline
  mov si,ax

  mov dx,03ceh
  mov ax,0005h
  out dx,ax
  mov ax,plane
  mov ah,al
  mov al,04h
  out dx,ax
  mov cx,40   { 40 words = 80 bytes = 1 scan line }
  les di,address
  rep movsw
  mov ax,1005h
  out dx,ax
  mov ax,0004h
  out dx,ax

  mov ds,bx

end;

{-----------------------------------------------------------------------}

procedure flushit;

begin

  blockwrite(fz,buf,bi);
  bi:=0;

end;

{-----------------------------------------------------------------------}

procedure write_pcx(fn : nt; var fl : boolean);

{
   Saves windowed region into a RLE .PCX file.
   *** Compression idea:

      Each scanline is decomposed into a continous stream of 4 bit planes
      = 80 bytes * 4 = 320 bytes.  This stream is checked for consecutive
      byte patterns.  Up to 63 ($3F) values can be compressed into two bytes
      forming a RLE code of $C0 or'ed with the run length (up to $3F). This
      forms the first byte, usually ($FF = 1100 0000 | 0011 1111).  The
      next byte is the actual data byte.  If a single byte is encountered
      it is written to the file as simply itself UNLESS the top two
      bits are set (mask $C0) which would indicate a RLE.  If this is
      the case then a special RLE of length 1 must be written in the
      general RLE form.  Here this would be ($C0 | 01 => $C1 followed
      by the data byte).  I compress each scan line separately thus ending
      possible RLE's at the end of each scan line.  This seems to be
      accepted practice.

      Bren Sessions, June 17, 1995
}

var header : header_rec;
    pal    : pal_rec;
    i,j,y  : integer;
    data   : array[0..319] of byte;   { (4 bitplanes) * (scan line=80 bytes) }
    dta    : byte;
    index  : integer;
    count  : integer;

begin

  if fn='' then   { exit if no filename given }
  begin
    fl:=false;
    exit;
  end;

  fillchar(header,sizeof(header),#0);
  with header do
  begin
    pcx_id  :=  10;
    pcx_ver :=   5;
    encode  :=   1;
    bpp     :=   1;
    left    :=   0;
    top     :=   0;
    right   := 639;
    bott    := 479;
    xres    :=   0;
    yres    :=   0;

    get_pal(pal);
    for i:=0 to 15 do
    begin
      if fl then j:=rm[i]
      else j:=i;
      rgb[i,0]:=pal[j,0] shl 2;
      rgb[i,1]:=pal[j,1] shl 2;
      rgb[i,2]:=pal[j,2] shl 2;
    end;

    resv    :=   0;
    bplanes :=   4;
    bpl     :=  80;   { bytes per line }
    ptype   :=   1;
    xsz     := 640;
    ysz     := 480;
  end;

  assign(fz,fn);
  {$i-} rewrite(fz,1); {$i+}
  if ioresult<>0 then
  begin
    writeln('Disk error.');
    fl:=false;
    exit;
  end;
  blockwrite(fz,header,sizeof(header));

  bi:=0;   { buffer index }

  for y:=0 to 479 do
  begin

    get_bitplane_at_scan_line(0,y,@data[0]);
    get_bitplane_at_scan_line(1,y,@data[80]);
    get_bitplane_at_scan_line(2,y,@data[160]);
    get_bitplane_at_scan_line(3,y,@data[240]);

    index:=0;

    repeat
      count:=0;
      dta:=data[index];
      repeat
        inc(index); inc(count);

        if count>$3F then
        begin
          if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=$FF;
          if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=dta;
          count:=1;
        end;
      until (index>319) or (data[index]<>dta);

      if count>1 then
      begin
        if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=$C0 or count;
        if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=dta;
      end
      else
      begin
        if (dta and $C0)=$C0 then
        begin
          if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=$C1;
          if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=dta;
        end
        else
        begin
          if bi=BUFSIZE then flushit; inc(bi); buf[bi]:=dta;
        end;
      end;

    until index>319;
  end;

  if bi>0 then flushit;

  close(fz);

  fl:=true;

end;

{-----------------------------------------------------------------------}

end.
